# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-06-27 14:49
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0009_auto_20160623_1521'),
        ('budget', '0011_alter_print_run_date'),
    ]

    operations = [
    ]



# Imports from python.  # NOQA
from datetime import datetime
from unittest import skipUnless


# Imports from Django.
from django.contrib.auth.models import AnonymousUser
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.test import SimpleTestCase
from django.test import TestCase
from django.utils.timezone import make_aware


# Imports from other dependencies.
from psycopg2.extras import DateTimeTZRange
import six


# Imports from budget.
from budget.models import Change
from budget.models import Headline
from budget.models import HeadlineVote
from budget.models import Item
from budget.models import Package
from budget.models import PrintPublication
from budget.models import PrintSection
from budget.tests.utils import budget_db
from budget.tests.utils import item_factory
from budget.tests.utils import package_factory
from budget.tests.utils import separate_db
from budget.tests.utils import staff_factory


__all__ = [
    'PrintPublicationTestCase',
    'PrintSectionTestCase',
    'HeadlineTestCase',
    'HeadlineVoteTestCase',
    'PackageTestCase',
    'ItemTestCase',
    'ChangeTestCase',
    'ChangeQuerysetTestCase',
    'CreationTrailTestCase',
]


class PrintPublicationTestCase(SimpleTestCase):
    def test_str(self):
        """Verify inactive suffix on publication name.

        Should return publication name with (inactive) appended when the
        publication is not active.
        """
        pub = PrintPublication(name='TDMN', slug='tdmn', priority=1)
        self.assertEqual(str(pub), 'TDMN')

        pub.is_active = False
        self.assertEqual(str(pub), 'TDMN (inactive)')


class PrintSectionTestCase(TestCase):
    multi_db = separate_db

    @classmethod
    def setUpTestData(cls):
        cls.pub = PrintPublication.objects.create(name='TDMN')

    def test_str_active(self):
        """Verify string method for PrintSection model.

        The method should return section name and pub name.
        """
        section = PrintSection.objects.create(name='News',
                                              publication=self.pub)
        self.assertEqual(str(section), 'TDMN > News')

    def test_str_section_inactive(self):
        """Verify string method behavior for inactive sections.

        When section is inactive, string representation should include
        a flag that it's inactive.
        """
        section = PrintSection.objects.create(name='News', is_active=False,
                                              publication=self.pub)
        self.assertEqual(str(section), 'TDMN > News (inactive)')

    def test_str_publication_inactive(self):
        """Verify string method behavior for inactive publications.

        When publication is inactive, string representation should
        include a flag that it's inactive.
        """
        self.pub.is_active = False
        section = PrintSection.objects.create(name='News',
                                              publication=self.pub)
        self.assertEqual(str(section), 'TDMN > News (inactive)')


class HeadlineTestCase(TestCase):
    multi_db = separate_db

    def test_str(self):
        """Verify string method for Headline model.

        The method should return the headline's text.
        """
        hed = Headline(
            package=package_factory(slug_key='test-hed-str'),
            text='A headline'
        )
        self.assertEqual(str(hed), 'A headline')

    def test_total_votes(self):
        """Verify total_votes() method behavior.

        The method should return latest total for headline votes.
        """
        hed = Headline.objects.create(
            package=package_factory(slug_key='test-hed-tv'),
            text='A headline'
        )
        for _ in range(5):
            HeadlineVote.objects.create(headline=hed, voter='a%s@b.com' % _)
        self.assertEqual(hed.total_votes(), 5)


class HeadlineVoteTestCase(TestCase):
    multi_db = separate_db

    def test_str(self):
        """Verify string method for HeadlineVote model.

        The method should include voter and vote information.
        """
        hed = Headline(
            package=package_factory(slug_key='test-hedvote-str'),
            text='A headline'
        )
        vote = HeadlineVote(headline=hed, voter='a@b.com')
        self.assertEqual(str(vote), 'a@b.com for A headline')


class PackageTestCase(TestCase):
    multi_db = separate_db

    def test_str(self):
        """Verify string method for Package model.

        The method should return generated slug information.
        """
        slug_key = 'my-test'
        package = package_factory(slug_key=slug_key)
        self.assertEqual(str(package), 'hub.{}.050115'.format(slug_key))

    def test_package_slug_creation(self):
        """Verify behavior of Package's 'full_slug' model property.

        This property should return a Package's 'slug_key'.
        """
        slug_key = 'test-key-one'
        package = package_factory(slug_key=slug_key)
        self.assertEqual(
            package.full_slug,
            'hub.{}.050115'.format(slug_key)
        )

    def test_publish_date_resolution_inference(self):
        """Verify inference of publish date based on date range length.

        The inferring method should treat ranges greater than 27 days as
        month (m) resolution, ranges between 6 and 27 days as week (w)
        resolution, ranges between 23 hours and 6 days as day (d)
        resolution and ranges less than 23 hours as time (t) resolution.
        """
        start_date = datetime(2015, 5, 1)

        staff_factory()

        time_resolution = Package.objects.create(
            publish_date=DateTimeTZRange(lower=start_date,
                                         upper=datetime(2015, 5, 1, 0, 1)),
            hub='hub')
        self.assertEqual(time_resolution.publish_date_resolution(), 't')

        day_resolution = Package.objects.create(
            publish_date=DateTimeTZRange(lower=start_date,
                                         upper=datetime(2015, 5, 2)),
            hub='hub')
        self.assertEqual(day_resolution.publish_date_resolution(), 'd')

        week_resolution = Package.objects.create(
            publish_date=DateTimeTZRange(lower=start_date,
                                         upper=datetime(2015, 5, 8)),
            hub='hub')
        self.assertEqual(week_resolution.publish_date_resolution(), 'w')

        month_resolution = Package.objects.create(
            publish_date=DateTimeTZRange(lower=start_date,
                                         upper=datetime(2015, 6, 1)),
            hub='hub')
        self.assertEqual(month_resolution.publish_date_resolution(), 'm')


class ItemTestCase(TestCase):
    multi_db = separate_db

    def test_primary_item_slug(self):
        """Primary item's slug should be None."""
        package = package_factory(slug_key='test-item-pis')
        primary_item = item_factory(primary_for=package)
        self.assertIsNone(primary_item.full_slug)

    def test_additional_item_slug(self):
        """Verify additional-item slug behavior.

        An additional item's slug should be the concatenation of its
        parent package's slug and the individual item's slug key.
        """
        slug_key = 'my-test-content'
        package = package_factory(slug_key='test-item-ais')
        additional_item = item_factory(
            additional_for=package,
            slug_key=slug_key
        )
        self.assertEqual(
            additional_item.full_slug,
            '{}.{}'.format(package.full_slug, slug_key)
        )

    def test_duplicate_additional_item_slug(self):
        """Verify protection against duplicate additional-item slugs.

        Model should throw an IntegrityError when saving multiple
        additional items with the same slug key to a package.
        """
        slug_key = 'my-test-content'
        package = package_factory(slug_key='test-item-dais')
        item_factory(
            additional_for=package,
            slug_key=slug_key
        )
        self.assertEqual(package.additional_content.count(), 1)
        self.assertEqual(
            package.additional_content.all()[0].slug_key,
            slug_key
        )
        with self.assertRaises(ValidationError):
            next_item = Item(
                slug_key=slug_key,
                type='text',
                editors=[{'email': 'e@d.com'}],
                authors=[{'email': 'a@d.com'}],
                budget_line='Budget line',
                primary_for_package=None,
                additional_for_package=package
            )
            next_item.validate_unique()

    def test_generate_date_slug_no_day(self):
        """Verify month and week date-range formatting behavior.

        If publish_date doesn't have at least day precision, day of
        month should be excluded from the date slug.
        """
        week_resolution = package_factory(
            publish_date_lower=make_aware(datetime(2015, 5, 1)),
            publish_date_upper=make_aware(datetime(2015, 5, 8)),
            slug_key='test-item-gdsnd'
        )
        self.assertEqual(week_resolution.slugified_date, '05--15')

    def test_generate_date_slug_with_day(self):
        """Verify precise-date date-range formatting behavior.

        If publish_date is day- / datetime-precise, generate a date slug
        with format 'MMDDYY'.
        """
        day_resolution = package_factory(
            publish_date_lower=make_aware(datetime(2015, 5, 1)),
            publish_date_upper=make_aware(datetime(2015, 5, 2)),
            slug_key='test-item-gdswd'
        )
        self.assertEqual(day_resolution.slugified_date, '050115')

        time_resolution = package_factory(
            publish_date_lower=make_aware(datetime(2015, 5, 1, 11)),
            publish_date_upper=make_aware(datetime(2015, 5, 2, 11, 1)),
            slug_key='test-item-gdswdat'
        )
        self.assertEqual(time_resolution.slugified_date, '050115')

    def test_primary_or_additional_required(self):
        """Verify items have no less than one link to the Package model.

        Raise ValidationError on an item if neither its primary nor its
        additional Package ForeignKeys have values.
        """
        with six.assertRaisesRegex(
            self,
            ValidationError,
            "Items must be connected to a package"
        ):
            item = Item(slug_key='sluggy-slug', type='text',
                        budget_line='Budget line')
            item.clean()

    def test_primary_and_additional_raise_error(self):
        """Verify mutual exclusivity of primary- & additional-type FKs.

        Raise ValidationError if an item has simultaneous values set for
        both primary and additional package.
        """
        with six.assertRaisesRegex(
            self,
            ValidationError,
            "Items cannot be connected to a package as both"
        ):
            package_one = package_factory(slug_key='test-item-tpaare1')
            package_two = package_factory(slug_key='test-item-tpaare2')
            item = Item(slug_key='sluggy-slug', type='text',
                        budget_line='Budget line',
                        primary_for_package=package_one,
                        additional_for_package=package_two)
            item.clean()


class ChangeTestCase(TestCase):
    multi_db = separate_db

    @classmethod
    def setUpTestData(cls):
        cls.item = item_factory(
            primary_for=package_factory(slug_key='test-change')
        )
        cls.user = User.objects.create_user('user')
        cls.change = Change.objects.create(
            by=cls.user.pk,
            package=cls.item.primary_for_package,
            item_content_type='item',
            item_id=cls.item.pk
        )

    def test_str(self):
        """Verify string method for Change model.

        The method should include user, package & timestamp information.
        """
        self.assertEqual(
            str(self.change),
            'User {} changed {} at {}.'.format(
                self.user.pk,
                str(self.item.primary_for_package),
                self.change.at
            )
        )

    def test_item(self):
        """This property should load the change's item instance."""
        self.assertEqual(self.change.item, self.item)

    def test_item_missing(self):
        """If Item model is missing, 'self.item' should be None."""
        self.item.delete()
        change = Change.objects.get(pk=self.change.pk)
        self.assertIsNone(change.item)

    def test_item_cache(self):
        """Verify multiple-reference behavior on 'Change.item' call.

        Subsequent calls after the first call to a Change model's item
        property should load from model's cache, not from additional
        queries to the database.
        """
        change = Change.objects.get(pk=self.change.pk)

        with self.assertNumQueries(1, using=budget_db):
            change.item

        with self.assertNumQueries(0, using=budget_db):
            change.item

    def test_user(self):
        """User property should load the change's user instance."""
        self.assertEqual(self.change.by_user, self.user)

    def test_user_cache(self):
        """Verify multiple-reference behavior on 'Change.by_user' call.

        Subsequent calls after the first call to a Change model's user
        property should load from model's cache, not from additional
        queries to the database.
        """
        change = Change.objects.get(pk=self.change.pk)

        with self.assertNumQueries(1, using='default'):
            change.by_user

        with self.assertNumQueries(0, using='default'):
            change.by_user

    def test_user_missing(self):
        """If User model is missing, 'self.by_user' should be None."""
        self.user.delete()
        change = Change.objects.get(pk=self.change.pk)
        self.assertIsNone(change.by_user)


class ChangeQuerysetTestCase(TestCase):
    multi_db = separate_db

    @classmethod
    def setUpTestData(cls):
        cls.user = User.objects.create_user('user')
        package = package_factory(slug_key='test-cqs')
        item = item_factory(primary_for=package)

        for _ in range(0, 5):
            Change.objects.create(
                by=cls.user.pk,
                package=package,
                item_content_type='package',
                item_id=package.pk
            )
            Change.objects.create(
                by=cls.user.pk,
                package=package,
                item_content_type='item',
                item_id=item.pk
            )

    @skipUnless(separate_db, 'Multiple databases required to test counts')
    def test_query_count(self):
        """Verify duplicate-removal behavior on fetch.

        Ensure that duplicate queries are eliminated when prefetching
        users and items.
        """
        # Control tests
        # Expect 21 total queries:
        # - 1 to get Changes
        # - 1x10 to get each related item
        # - 1x10 to get each related user
        with self.assertNumQueries(11, using='budget'), \
                self.assertNumQueries(10, using='default'):
            for _ in Change.objects.all().select_related('package'):
                _.item
                _.by_user

        # Test against both default and budget DBs because we're loading users
        # Expect four total queries:
        # - 1 to budget to get Changes
        # - 1 to budget to get related Items
        # - 1 to budget to get related Packages
        # - 1 to default to get related users
        with self.assertNumQueries(3, using='budget'), \
                self.assertNumQueries(1, using='default'):
            for _ in Change.objects.all().select_related('package').\
                    prefetch_related('by_user', 'item'):
                _.item
                _.by_user

    def test_empty_queryset(self):
        """Verify queryset's ability to return empty querysets."""
        self.assertEqual(len(Change.objects.filter(by=10000)), 0)

    def test_values_query(self):
        """Verify tolerance of non-model-instance returning queries."""
        user_ids = Change.objects.all().values('by')
        expect = [{'by': self.user.pk} for x in range(0, 10)]
        self.assertListEqual(list(user_ids), list(expect))


class CreationTrailTestCase(TestCase):
    """Tests for our abstract base class.

    This class adds 'created_by' and 'created_at' fields.

    All tests use the Package model, but this class is mixed into almost
    all budget models.
    """
    multi_db = separate_db

    def test_created_by_user(self):
        """Verify behavior of 'created_by_user' model property.

        This property should load the user model from the database.
        """
        user = User.objects.create_user('user')

        package = package_factory(slug_key='test-cr-trl-cbu')
        package.created_by = user.pk
        package.save()

        self.assertEqual(package.created_by_user, user)

    def test_created_by_user_none(self):
        """Verify 'created_by_user' behavior when user is null.

        In this case, the property should return AnonymousUser.
        """
        package = package_factory(slug_key='test-cr-trl-cbun')
        self.assertEqual(package.created_by_user, AnonymousUser)

    def test_created_by_user_missing(self):
        """Verify 'created_by_user' behavior when user does not exist.

        In this case, the property should return a value of None.
        """
        package = package_factory(slug_key='test-cr-trl-cbum')
        package.created_by = 5000
        package.save()

        self.assertEqual(package.created_by_user, None)



# Imports from Django.  # NOQA
from django.core.management.base import BaseCommand


# Imports from other dependencies.
from six.moves import input


# Imports from budget.
from budget.models import Package


class Command(BaseCommand):
    help = 'Deletes all packages without a primary content item'

    def add_arguments(self, parser):
        parser.add_argument(
            '--noinput', '--no-input',
            action='store_false', dest='interactive', default=True,
            help=(
                'Tells Django to NOT prompt the user for input of any kind. '
                'Packages without primary content items will be deleted '
                'without asking for confirmation.'
            ),
        )

    def handle(self, *args, **options):
        with_orphans = Package.objects.filter(primary_content__isnull=True)

        if not with_orphans:
            self.stdout.write(self.style.SUCCESS(
                'No packages without primary content items found.'))
            return

        deleted = 0

        for orphan in with_orphans:
            if options['interactive']:
                delete = False
                while delete is False:
                    msg = 'Delete package "%s"? [y/n] ' % orphan
                    answer = input(msg).lower()
                    if answer == 'y':
                        self.stdout.write('Deleting "%s".' % orphan)
                        orphan.delete()
                        deleted += 1
                        delete = True
                    elif answer == 'n':
                        self.stdout.write('Skipping "%s".' % orphan)
                        break
                    else:
                        self.stdout.write(self.style.WARNING(
                            'Answer with "y" to delete or "n" to skip.'))
            else:
                self.stdout.write('Deleting "%s"' % orphan)
                orphan.delete()
                deleted += 1

        self.stdout.write(self.style.SUCCESS(
            'Deleted %s package(s) without primary content items.' %
            len(with_orphans)))



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-03-28 18:47
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='package',
            name='headline_voting_open',
        ),
        migrations.AddField(
            model_name='package',
            name='headline_status',
            field=models.CharField(choices=[(b'drafting', b'Drafting'), (b'voting', b'Voting'), (b'finalized', b'Finalized')], default=b'drafting', max_length=9),
        ),
        migrations.AlterField(
            model_name='headlinevote',
            name='headline',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='headline_vote', to='budget.Headline'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-08-20 23:30
from __future__ import unicode_literals


from django.db import IntegrityError
from django.db import migrations


from budget.utils import SLUG_MATCH_RE


def forward_func(apps, schema_editor):
    """TK."""
    Item = apps.get_model('budget', 'Item')

    for item in Item.objects.all():
        matched_slug = SLUG_MATCH_RE.match(item.slug)

        if matched_slug:
            slug_parts = matched_slug.groupdict()

            if item.primary_for_package:
                if slug_parts.get('suffix_letter', ''):
                    primary_slug = '{}--{}'.format(
                        slug_parts['slug_key'],
                        slug_parts['suffix_letter']
                    )
                else:
                    primary_slug = slug_parts['slug_key']

                package = item.primary_for_package

                preserved_slug = 'Migrated -- original slug "{}"'.format(
                    item.slug
                )

                package.slug_key = primary_slug
                package.notes = '{}\n{}'.format(package.notes, preserved_slug)
                package.save()

                item.slug_key = None

            elif item.additional_for_package:
                if slug_parts.get('additional_slug') is not None:
                    additional_slug = slug_parts['additional_slug']
                else:
                    raise IntegrityError(
                        ' '.join([
                            'The additional content item "{}" doesn\'t',
                            'have a slug.',
                            'Halting migration.',
                        ]).format(str(item))
                    )

                item.slug_key = additional_slug
            else:
                raise IntegrityError(
                    ' '.join([
                        'The budget item "{}" isn\'t connected as either the',
                        'primary piece of content in a package or as an',
                        'additional piece of content.',
                        'Halting migration.',
                        # ]).format(str(item))
                    ]).format(str(item.id))
                )

        else:
            raise IntegrityError(
                ' '.join([
                    'Couldn\'t parse slug from budget item "{}".',
                    'Halting migration.',
                ]).format(str(item))
            )

        item.save()


def backward_func(apps, schema_editor):
    """TK."""
    Item = apps.get_model('budget', 'Item')

    for item in Item.objects.all():
        if item.primary_for_package:
            package = item.primary_for_package

            slug_and_suffix = package.slug_key.split('--')

            if len(slug_and_suffix) > 1:
                suffix = slug_and_suffix[-1]

                item_slug = '{}.{}.{}{}'.format(
                    package.hub,
                    '--'.join(slug_and_suffix[:-1]),
                    package.slugified_date,
                    suffix
                )
            else:
                item_slug = '{}.{}.{}'.format(
                    package.hub,
                    package.slug_key,
                    package.slugified_date,
                )

            item.slug_key = package.slug_key

            package.slug_key = ''
            package.save()

        elif item.additional_for_package:
            package = item.additional_for_package

            slug_and_suffix = package.slug_key.split('--')

            if len(slug_and_suffix) > 1:
                suffix = slug_and_suffix[-1]

                item_slug = '{}.{}.{}{}.{}'.format(
                    package.hub,
                    '--'.join(slug_and_suffix[:-1]),
                    package.slugified_date,
                    suffix,
                    item.slug_key
                )
            else:
                item_slug = '{}.{}.{}.{}'.format(
                    package.hub,
                    package.slug_key,
                    package.slugified_date,
                    item.slug_key
                )

        else:
            raise IntegrityError(
                ' '.join([
                    'The budget item "{}" isn\'t connected as either the',
                    'primary piece of content in a package or as an',
                    'additional piece of content.',
                    'Halting migration.'
                ])
            )
        item.slug = item_slug

        item.save()


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0028_auto_20170820_1830'),
    ]

    operations = [
        migrations.RunPython(forward_func, backward_func)
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-03-28 00:42
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Headline',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('text', models.TextField()),
                ('winner', models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name='HeadlineVote',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('voter', models.EmailField(max_length=254)),
                ('timestamp', models.DateTimeField(auto_now_add=True)),
                ('headline', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='budget.Headline')),
            ],
        ),
        migrations.CreateModel(
            name='Item',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', models.CharField(max_length=100, unique=True)),
                ('type', models.CharField(choices=[(b'text', b'Text'), (b'photo', b'Photo'), (b'video', b'Video'), (b'audio', b'Audio'), (b'graphic', b'Graphic'), (b'interactive', b'Interactive'), (b'other', b'Other')], default=b'text', max_length=25)),
                ('editors', django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True)),
                ('authors', django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True)),
                ('subscribers', django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True)),
                ('is_ready', models.BooleanField(default=False)),
                ('budget_line', models.TextField()),
                ('length', models.PositiveIntegerField(blank=True, help_text=b'How long is this article?', null=True, verbose_name=b'Length (in inches)')),
            ],
            options={
                'ordering': ['slug'],
            },
        ),
        migrations.CreateModel(
            name='Package',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('vertical', models.CharField(blank=True, max_length=100, null=True)),
                ('hub', models.CharField(max_length=100)),
                ('url', models.URLField(blank=True, null=True)),
                ('print_issue', models.DateField(blank=True, null=True)),
                ('print_pages', models.CharField(blank=True, max_length=50, null=True)),
                ('pub_date', models.DateTimeField()),
                ('pub_date_resolution', models.CharField(choices=[(b'm', b'Month'), (b'w', b'Week'), (b'd', b'Day'), (b't', b'Time')], default=b'd', max_length=1)),
                ('notes', models.TextField(blank=True, null=True)),
                ('created_by', models.EmailField(blank=True, max_length=254, null=True)),
                ('last_changed_by', models.EmailField(blank=True, max_length=254, null=True)),
                ('serialized_object', django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True)),
                ('headline_voting_open', models.BooleanField(default=False)),
            ],
            options={
                'ordering': ['pub_date', 'primary_content__slug'],
            },
        ),
        migrations.AddField(
            model_name='item',
            name='additional_for_package',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='additional_content', to='budget.Package'),
        ),
        migrations.AddField(
            model_name='item',
            name='primary_for_package',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='primary_content', to='budget.Package'),
        ),
        migrations.AddField(
            model_name='headline',
            name='package',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='headlines', to='budget.Package'),
        ),
        migrations.AlterUniqueTogether(
            name='headlinevote',
            unique_together=set([('headline', 'voter')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-06-27 18:12
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0015_auto_20160627_1308'),
    ]

    operations = [
        migrations.RenameField(
            model_name='package',
            old_name='url',
            new_name='published_url',
        ),
        migrations.AlterField(
            model_name='headline',
            name='package',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='headline_candidates', to='budget.Package'),
        ),
    ]



# Imports from Django.  # NOQA
from django.conf.urls import include
from django.conf.urls import url


# Imports from other dependencies.
from rest_framework import routers


# Imports from budget.
from budget.views import ConfigView
from budget.views import ContentPlacementViewSet
from budget.views import ExceptionView
from budget.views import HeadlineViewSet
from budget.views import HeadlineVoteViewSet
from budget.views import ItemViewSet
from budget.views import MainBudgetView
from budget.views import PackageViewSet
from budget.views import PrintPublicationViewSet
from budget.views import UserEditingPackageUpdateView
from budget.views import UserExitedPackageUpdateView


app_name = 'budget'


router = routers.DefaultRouter()
router.register(r'packages', PackageViewSet)
router.register(r'items', ItemViewSet)
router.register(r'headlines', HeadlineViewSet)
router.register(r'print-publications', PrintPublicationViewSet)
router.register(r'headline-votes', HeadlineVoteViewSet)
router.register(r'content-placements', ContentPlacementViewSet)


urlpatterns = [
    url(r'^api/', include(router.urls)),

    url(r'^exception/$', ExceptionView.as_view()),

    # Add an extra custom route here because the built-in REST RegEx won't
    # match our slugs because of the double-dots
    url(r'^api/user-activity/', include([
        url(r'^editing-(?P<package_id>[\d]+)/$',
            UserEditingPackageUpdateView.as_view(),
            name='user-editing-package'),
        url(r'^exited-(?P<package_id>[\d]+)/$',
            UserExitedPackageUpdateView.as_view(),
            name='user-exited-package')
    ])),

    url(r'^api/packages/(?P<pk>[\d]+)/$', PackageViewSet.as_view({
        'get': 'retrieve',
        'put': 'update',
        'patch': 'partial_update',
        'delete': 'destroy'
    })),
    url(r'^api/packages/(?P<raw_slug>[-\w.]+)/$', PackageViewSet.as_view({
        'get': 'retrieve',
        'put': 'update',
        'patch': 'partial_update',
        'delete': 'destroy'
    })),

    url(r'^config/$', ConfigView.as_view(), name='config'),

    url(r'^(?P<path>.*)$', MainBudgetView.as_view(), name='main-budget-view'),
]



# Imports from Django.  # NOQA
from django.contrib import admin


# Imports from other dependencies.
from adminsortable2.admin import SortableInlineAdminMixin


# Imports from budget.
from budget.models import Change
from budget.models import ContentPlacement
from budget.models import Headline
from budget.models import HeadlineVote
from budget.models import Item
from budget.models import Package
from budget.models import PrintPublication
from budget.models import PrintSection


admin.site.register(Headline)
admin.site.register(HeadlineVote)


class ContentPlacementInline(admin.StackedInline):
    model = ContentPlacement
    extra = 0
    fields = (
        'destination',
        ('run_date', 'external_slug',),
        ('placement_types', 'placement_details',),
        'is_finalized',
    )


@admin.register(Package)
class PackageAdmin(admin.ModelAdmin):
    """Admin config for budget packages."""
    list_display = ('full_slug', 'hub', 'scheduled_publish_date')
    list_filter = ('hub', 'vertical',)
    ordering = ('-id',)
    fieldsets = (
        (None, {
            'fields': ('vertical', 'hub')
        }),
        ('Digital publishing info', {
            'fields': ('publish_date', 'published_url', 'notes')
        }),
        # ('Print publishing info', {
        #     'fields': (
        #         'print_section',
        #         'print_system_slug',
        #         'print_run_date',
        #         'print_placements',
        #         'is_print_placement_finalized'
        #     )
        # }),
        ('Headlines', {
            'fields': ('headline_status',)
        }),
        ('Debugging info', {
            'fields': ('created_by', 'last_changed_by_old',)
        })
    )
    inlines = [
        ContentPlacementInline
    ]

    def scheduled_publish_date(self, obj):
        return '%s (%s)' % (
            obj.publish_date.lower.strftime('%B %-d, %Y'),
            obj.publish_date_resolution()
        )


@admin.register(Item)
class ItemAdmin(admin.ModelAdmin):
    """Admin config for budget items."""
    fieldsets = (
        (None, {
            'fields': ('slug_key',)
        }),
        ('Related package', {
            'fields': ('primary_for_package', 'additional_for_package',)
        }),
        ('Budget info', {
            'fields': ('type', 'budget_line', 'length', 'is_ready')
        }),
        ('Authors', {
            'fields': ('authors', 'editors')
        })
    )
    list_display = ('full_slug', 'type', 'is_primary',)
    list_filter = ('type',)
    ordering = ('-id',)

    def get_queryset(self, request):
        qs = super(ItemAdmin, self).get_queryset(request)
        return qs.select_related(
            'primary_for_package',
            'additional_for_package'
        )

    def is_primary(self, obj):
        if obj.primary_for_package:
            return True
        return False
    is_primary.boolean = True


class PrintSectionInline(SortableInlineAdminMixin, admin.TabularInline):
    model = PrintSection
    extra = 0


@admin.register(PrintPublication)
class PrintPublicationAdmin(admin.ModelAdmin):
    """Admin config for print publications."""
    fieldsets = (
        (None, {
            'fields': (
                'name',
                'priority',
                'is_active'
            )
        }),
        ("Don't touch unless you know what you're doing", {
            'classes': ('collapse',),
            'fields': ('slug',),
        }),
    )
    inlines = [PrintSectionInline]
    prepopulated_fields = {
        'slug': ('name',)
    }


@admin.register(Change)
class ChangeAdmin(admin.ModelAdmin):
    """Admin config for budget-item change records."""
    list_display = ('pk', 'by_user', 'at', 'package', 'item',
                    'item_content_type', 'item_id',)
    list_filter = ('item_content_type',)
    date_hierarchy = 'at'

    def get_queryset(self, request):
        return super(ChangeAdmin, self)\
            .get_queryset(request)\
            .select_related('package', 'package__primary_content')\
            .prefetch_related('by_user', 'item')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-08-18 16:23
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0025_auto_20170817_1931'),
    ]

    operations = [
        migrations.AlterField(
            model_name='item',
            name='slug',
            field=models.CharField(db_index=True, editable=False, max_length=100),
        ),
    ]



# Imports from python.  # NOQA
from datetime import datetime
from datetime import timedelta
import json


# Imports from Django.
from django.contrib.auth.models import User
# from django.test import override_settings
from django.test import TestCase
from django.urls import reverse
from django.utils.timezone import make_aware


# Imports from other dependencies.
from rest_framework.test import APIClient


# Imports from budget.
from budget.models import Change
# from budget.models import Headline
# from budget.models import HeadlineVote
from budget.models import Item
from budget.models import Package
from budget.models import PrintPublication
from budget.models import PrintSection
from budget.tests.utils import budget_db
from budget.tests.utils import get_timestamp_for_date
from budget.tests.utils import item_factory
from budget.tests.utils import package_factory
from budget.tests.utils import separate_db
from budget.tests.utils import staff_factory
from budget.tests.utils import tz

__all__ = [
    'CommonAPITestCase',
    'HeadlinesAPITestCase',
    'ItemsAPITestCase',
    'PackagesAPITestCase',
    'PrintPublicationAPITestCase',
    'AuditTrailApiMixinTestCase',
]


PACKAGES_API_ENDPOINT = '/api/packages/'
PUBLICATION_API_ENDPOINT = '/api/print-publications/'
ITEMS_API_ENDPOINT = reverse('budget:item-list')


class AuthedApiTestMixin(TestCase):
    """Mixin to simulate API request client.

    Store an authed API request client on the instance, which we can
    later use to test our API.
    """
    def setUp(self):
        super(AuthedApiTestMixin, self).setUp()
        user = User.objects.create_user('user')
        self.client = APIClient()
        self.client.force_authenticate(user=user)


# @override_settings(ROOT_URLCONF='budget.test_urls')
class CommonAPITestCase(AuthedApiTestMixin, TestCase):
    """Tests for shared (or potentially shareable) logic from our API.

    Includes tests for custom serializers, filters, etc.
    """
    multi_db = separate_db

    @classmethod
    def setUpTestData(cls):
        for _ in range(5):
            package_slug = 'test-package-{}'.format(_)
            item_factory(primary_for=package_factory(slug_key=package_slug))

    def test_invalid_publish_date_filter(self):
        """Verify handling of invalid publish_date filter values.

        When invalid params are passed in the publish_date filter an
        empty queryset should be returned.

        Tests the packages API endoint, but really we're checking all
        endpoints that use our shared parsing/validation decorator
        (budget.filters.daterange_filter_decorator).
        """
        # Run a valid query just to be 1000% sure we're not just getting back
        # empty queries
        query = {'format': 'json', 'publish_date': '2015-05-01,2015-05-02'}
        response = self.client.get(PACKAGES_API_ENDPOINT, data=query)
        self.assertEqual(json.loads(response.content)['count'], 5)

        # Missing date
        query = {'format': 'json', 'publish_date': '2015-05-01'}
        response = self.client.get(PACKAGES_API_ENDPOINT, data=query)
        self.assertEqual(json.loads(response.content)['count'], 0)

        # Invalid date format
        query = {'format': 'json', 'publish_date': 'not-a-date,2015-05-01'}
        response = self.client.get(PACKAGES_API_ENDPOINT, data=query)
        self.assertEqual(json.loads(response.content)['count'], 0)

        # Lower bound greater than upper
        query = {'format': 'json', 'publish_date': '2016-05-01,2015-05-01'}
        response = self.client.get(PACKAGES_API_ENDPOINT, data=query)
        self.assertEqual(json.loads(response.content)['count'], 0)

    def test_publish_date_tz_handling(self):
        """Verify date-query timezone handling.

        Date queries should perform time zone-aware queries in
        local time.
        """
        query = {'format': 'json', 'publish_date': '2015-04-30,2015-05-01'}
        response = self.client.get(PACKAGES_API_ENDPOINT, data=query)
        self.assertEqual(json.loads(response.content)['count'], 0)

        query = {'format': 'json', 'publish_date': '2015-05-02,2015-05-03'}
        response = self.client.get(PACKAGES_API_ENDPOINT, data=query)
        self.assertEqual(json.loads(response.content)['count'], 0)

    def test_read_from_daterange_field(self):
        """Verify range-field serialization behavior.

        Range field should serialize as two ISO-8601 dates in an array.
        """
        package = package_factory(slug_key='test-daterng-read')
        package_url = '%s%s/' % (PACKAGES_API_ENDPOINT, package.pk,)

        response = self.client.get(package_url, data={'format': 'json'})
        json_response = json.loads(response.content)
        expect = [
            get_timestamp_for_date('2015-05-01'),
            get_timestamp_for_date('2015-05-02'),
        ]
        self.assertEqual(json_response['publishDate'], expect)

    def test_write_to_daterange_field(self):
        """Verify parsing and storage of date-range values.

        Valid ISO-8601 dates passed in a range field should be parsed
        and stored.
        """
        package = package_factory(slug_key='test-daterng-write-1')
        package_url = '%s%s/?format=json' % (
            PACKAGES_API_ENDPOINT, package.pk,)

        update = {
            'publishDate': ['2015-05-03T00:00:00Z', '2015-05-04T00:00:00Z']
        }
        response = self.client.patch(package_url, update, format='json')
        self.assertEqual(response.status_code, 200)

        json_response = json.loads(response.content)
        self.assertEqual(json_response['publishDate'], update['publishDate'])

    def test_write_invalid_date_format_to_daterange_field(self):
        """Invalid ISO-8601 dates throw a validation error"""
        package = package_factory(slug_key='test-daterng-write-2')
        package_url = '%s%s/?format=json' % (
            PACKAGES_API_ENDPOINT, package.pk,)

        update = {
            'publishDate': ['not-a-date', '2015-05-04T00:00:00Z']
        }
        response = self.client.patch(package_url, update, format='json')
        self.assertEqual(response.status_code, 400)

        error = {
            'publishDate': [
                ' '.join([
                    'Datetime has wrong format.',
                    'Use one of these formats instead:',
                    'YYYY-MM-DDThh:mm[:ss[.uuuuuu]][+HH:MM|-HH:MM|Z].',
                ])
            ]
        }
        self.assertJSONEqual(response.content, error)

    def test_write_invalid_date_count_to_daterange_field(self):
        """Verify invalid-length arrays fail date-range validation.

        Should raise ValidationError when the date-range value is not
        passed enough date values.
        """
        package = package_factory(slug_key='test-daterng-write-3')
        package_url = '%s%s/?format=json' % (
            PACKAGES_API_ENDPOINT, package.pk,)

        update = {
            'publishDate': ['2015-05-02T00:00:00Z']
        }
        response = self.client.patch(package_url, update, format='json')
        self.assertEqual(response.status_code, 400)

        error = {
            'publishDate': [
                'Incorrect format. Expected an Array with two items.',
            ]
        }
        self.assertJSONEqual(response.content, error)

    def test_write_invalid_date_bounds_to_daterange_field(self):
        """Verify date-range part ordering behavior.

        Passing a date range where the lower is greater than the upper
        raises an error.
        """
        package = package_factory(slug_key='test-daterng-write-4')
        package_url = '%s%s/?format=json' % (
            PACKAGES_API_ENDPOINT, package.pk,)

        update = {
            'publishDate': [
                get_timestamp_for_date('2015-05-03'),
                get_timestamp_for_date('2015-05-02'),
            ]
        }
        response = self.client.patch(package_url, update, format='json')
        self.assertEqual(response.status_code, 400)

        error = {
            'publishDate': [
                'The upper date bound must be greater than the lower bound.'
                ]
        }
        self.assertJSONEqual(response.content, error)

    def test_empty_json_serializer(self):
        """Verify null-value JSON serialization behavior.

        When serialized with our custom EmptyArrayJsonField, null values
        should be serialized as empty arrays.
        """
        # Don't use our factory, because we need None for authors/editors
        slug_key = 'test-json-serializer'
        item = Item.objects.create(
            slug_key='sluggy-slug',
            type='text',
            budget_line='Budget line',
            primary_for_package=package_factory(slug_key=slug_key)
        )
        item_url = '%s%s/?format=json' % (ITEMS_API_ENDPOINT, item.pk,)
        response = self.client.get(item_url)
        self.assertEqual(response.status_code, 200)
        json_response = json.loads(response.content)
        self.assertEqual(json_response['authors'], [])
        self.assertEqual(json_response['editors'], [])


class HeadlinesAPITestCase(TestCase):
    pass


# @override_settings(ROOT_URLCONF='budget.test_urls')
class ItemsAPITestCase(AuthedApiTestMixin, TestCase):
    multi_db = separate_db

    @classmethod
    def setUpTestData(cls):
        package_one = package_factory(slug_key='test-items-api-base')
        item_factory(primary_for=package_one)
        item_factory(additional_for=package_one, slug_key='addl1')
        # item_factory(additional_for=package_one)
        for _ in range(5):
            slug_key = 'test-items-api-{}'.format(_)
            item_factory(primary_for=package_factory(slug_key=slug_key))

    def test_query_count(self):
        """Only make two queries to budget DB for an item list view."""
        with self.assertNumQueries(2, using=budget_db):
            self.client.get(ITEMS_API_ENDPOINT, data={'format': 'json'})

    def test_person_filter(self):
        """Person filter should search both author & editor fields."""
        item_factory(
            author='a-person@b.com',
            primary_for=package_factory(slug_key='test-items-api-p1')
        )
        item_factory(
            editor='a-person@b.com',
            primary_for=package_factory(slug_key='test-items-api-p2')
        )
        response = self.client.get(ITEMS_API_ENDPOINT, data={
            'format': 'json',
            'person': 'a-person@b.com'
        })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json.loads(response.content)['count'], 2)

    def test_author_filter(self):
        """Author field should return items based on author's e-mail."""
        item_factory(
            author='an-author@b.com',
            primary_for=package_factory(slug_key='test-items-api-p3')
        )
        response = self.client.get(ITEMS_API_ENDPOINT, data={
            'format': 'json',
            'author': 'an-author@b.com'
        })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json.loads(response.content)['count'], 1)

    def test_editor_filter(self):
        """Editor field should return items based on editor's e-mail."""
        item_factory(
            editor='an-editor@b.com',
            primary_for=package_factory(slug_key='test-items-api-p4')
        )

        response = self.client.get(ITEMS_API_ENDPOINT, data={
            'format': 'json',
            'editor': 'an-editor@b.com'
        })

        self.assertEqual(response.status_code, 200)
        self.assertEqual(json.loads(response.content)['count'], 1)

    def test_date_filter(self):
        """Verify publish_date filter behavior.

        Publish date filter should return items based on primary and
        additional package's publish date.
        """
        package = package_factory(
            publish_date_lower=make_aware(datetime(2015, 7, 4)),
            publish_date_upper=make_aware(datetime(2015, 7, 5)),
            slug_key='test-items-api-d1'
        )
        item_factory(primary_for=package)
        item_factory(additional_for=package)
        response = self.client.get(ITEMS_API_ENDPOINT, data={
            'format': 'json',
            'publish_date': '2015-07-04,2015-07-05'
        })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json.loads(response.content)['count'], 2)


# @override_settings(ROOT_URLCONF='budget.test_urls')
class PackagesAPITestCase(AuthedApiTestMixin, TestCase):
    multi_db = separate_db

    @classmethod
    def setUpTestData(cls):
        # Generate packages on 5 consecutive days, so we can run date-bound
        # filtered queries
        for offset in range(5):
            days_to_add = offset + 1
            lower = tz.localize(datetime(2015, 5, 1)) + timedelta(
                days=days_to_add
            )
            upper = lower + timedelta(days=1)
            package = package_factory(
                publish_date_lower=lower,
                publish_date_upper=upper)
            item_factory(primary_for=package)

    def test_query_count(self):
        """Only make 4 queries to budget DB for a package list view."""
        with self.assertNumQueries(5, using=budget_db):
            self.client.get(PACKAGES_API_ENDPOINT, data={'format': 'json'})

    def test_publish_date_filter(self):
        """Verify publish_date filter behavior.

        A comma-separated publish_date param should perform a range
        query on publish_date.
        """
        # We create one package per day in the setup, so number of days in
        # the range should equal number of packages
        query = {'format': 'json', 'publish_date': '2015-05-02,2015-05-03'}
        response = self.client.get(PACKAGES_API_ENDPOINT, data=query)
        self.assertEqual(json.loads(response.content)['count'], 1)

        query['publish_date'] = '2015-05-02,2015-05-05'
        response = self.client.get(PACKAGES_API_ENDPOINT, data=query)
        self.assertEqual(json.loads(response.content)['count'], 3)

    def test_get_by_slug(self):
        """The package detail view should allow GETing by slug"""
        package = package_factory(slug_key='test-pkg-api-gbs')

        url = '%s%s/' % (PACKAGES_API_ENDPOINT, package.full_slug)
        response = self.client.get(url, data={'format': 'json'})
        self.assertEqual(response.status_code, 200)

        json_response = json.loads(response.content)
        self.assertEqual(json_response['id'], package.id)

    def test_get_by_id(self):
        """Verify package API get-by-ID behavior.

        This is the traditional GET behavior for API detail views.
        """
        package = package_factory(slug_key='test-pkg-api-gbid')

        url = '%s%s/' % (PACKAGES_API_ENDPOINT, package.pk)
        response = self.client.get(url, data={'format': 'json'})
        self.assertEqual(response.status_code, 200)

        json_response = json.loads(response.content)
        self.assertEqual(json_response['id'], package.id)

    def test_sort_by_publish_date(self):
        """Verify package API 'publish_date' sorting behavior.

        When this sort option is specified, the API should sort first by
        a package's upper bound, then by the precision of its range.
        """
        # We already have four packages in the database, each of which are one
        # day long. We need to add two more here with > 1 day duration to test
        # the range precision sorting.
        package_factory(
            publish_date_lower=tz.localize(datetime(2015, 5, 2)),
            publish_date_upper=tz.localize(datetime(2015, 5, 5)),
            slug_key='test-pkg-api-sbpd1'
        )
        package_factory(
            publish_date_lower=tz.localize(datetime(2015, 5, 3)),
            publish_date_upper=tz.localize(datetime(2015, 5, 5)),
            slug_key='test-pkg-api-sbpd2'
        )

        response = self.client.get(PACKAGES_API_ENDPOINT, data={
            'format': 'json', 'ordering': 'publish_date'})

        # Get the publish dates of the returned packages and make sure they
        # align with what we expect
        returned_dates = [
            x['publishDate'] for x in json.loads(response.content)['results']
        ]
        expected_dates = [
            [
                get_timestamp_for_date('2015-05-06'),
                get_timestamp_for_date('2015-05-07'),
            ],
            [
                get_timestamp_for_date('2015-05-05'),
                get_timestamp_for_date('2015-05-06'),
            ],
            [
                get_timestamp_for_date('2015-05-04'),
                get_timestamp_for_date('2015-05-05'),
            ],
            [
                get_timestamp_for_date('2015-05-03'),
                get_timestamp_for_date('2015-05-05'),
            ],
            [
                get_timestamp_for_date('2015-05-02'),
                get_timestamp_for_date('2015-05-05'),
            ],
            [
                get_timestamp_for_date('2015-05-03'),
                get_timestamp_for_date('2015-05-04'),
            ],
            [
                get_timestamp_for_date('2015-05-02'),
                get_timestamp_for_date('2015-05-03'),
            ]
        ]
        self.assertListEqual(returned_dates, expected_dates)

    def test_default_sort(self):
        """By default, packages API should order by 'publish_date'."""
        # Add additional packages with > 1 day duration to test range precision
        package_factory(
            publish_date_lower=datetime(2015, 5, 1),
            publish_date_upper=datetime(2015, 5, 4),
            slug_key='test-pkg-api-dst1'
        )
        package_factory(
            publish_date_lower=datetime(2015, 5, 2),
            publish_date_upper=datetime(2015, 5, 4),
            slug_key='test-pkg-api-dst2'
        )

        response_with_sort = self.client.get(
            PACKAGES_API_ENDPOINT,
            data={
                'format': 'json',
                'ordering': 'publish_date',
            }
        )
        response_no_sort = self.client.get(
            PACKAGES_API_ENDPOINT,
            data={'format': 'json'}
        )

        ids_with_sort = ','.join([
            str(_['id'])
            for _ in json.loads(response_with_sort.content)['results']
        ])

        ids_no_sort = ','.join([
            str(_['id'])
            for _ in json.loads(response_no_sort.content)['results']
        ])

        self.assertEqual(ids_with_sort, ids_no_sort)

    def test_searches_budget_lines(self):
        """Verify search query behavior.

        Searches should find packages based on both primary & additional
        content budget lines.
        """
        package = package_factory(slug_key='test-pkg-api-bg-line')
        item_factory(primary_for=package, budget_line='abcdefghij')
        item_factory(additional_for=package, budget_line='klmnopqrst')

        response = self.client.get(PACKAGES_API_ENDPOINT, data={
            'format': 'json',
            'search': 'abcdefghij'
        })
        json_response = json.loads(response.content)
        self.assertEqual(json_response['count'], 1)
        self.assertEqual(json_response['results'][0]['id'], package.pk)

        response = self.client.get(PACKAGES_API_ENDPOINT, data={
            'format': 'json',
            'search': 'klmnopqrst'
        })
        json_response = json.loads(response.content)
        self.assertEqual(json_response['count'], 1)
        self.assertEqual(json_response['results'][0]['id'], package.pk)

    def test_searches_slugs(self):
        """Verify search query behavior.

        Searches should find packages based on both primary & additional
        content slugs.
        """
        package = package_factory(slug_key='test-pkg-api-slugs')
        item_factory(primary_for=package)
        additional_item = item_factory(
            additional_for=package,
            slug_key='klmnopqrst'
        )

        response = self.client.get(PACKAGES_API_ENDPOINT, data={
            'format': 'json',
            'search': package.full_slug
        })
        json_response = json.loads(response.content)
        self.assertEqual(json_response['count'], 1)
        self.assertEqual(json_response['results'][0]['id'], package.pk)

        response = self.client.get(PACKAGES_API_ENDPOINT, data={
            'format': 'json',
            'search': additional_item.full_slug
        })
        json_response = json.loads(response.content)
        self.assertEqual(json_response['count'], 1)
        self.assertEqual(json_response['results'][0]['id'], package.pk)

    def test_searches_slug_keys(self):
        """Verify search query behavior.

        Searches should find packages based on both primary & additional
        content slug keys.
        """
        package = package_factory(slug_key='test-pkg-api-ssk')
        item_factory(primary_for=package)
        item_factory(additional_for=package, slug_key='klmnopqrst')

        response = self.client.get(PACKAGES_API_ENDPOINT, data={
            'format': 'json',
            'search': 'test-pkg-api-ssk'
        })
        json_response = json.loads(response.content)
        self.assertEqual(json_response['count'], 1)
        self.assertEqual(json_response['results'][0]['id'], package.pk)

        response = self.client.get(PACKAGES_API_ENDPOINT, data={
            'format': 'json',
            'search': 'klmnopqrst'
        })
        json_response = json.loads(response.content)
        self.assertEqual(json_response['count'], 1)
        self.assertEqual(json_response['results'][0]['id'], package.pk)

    def test_person_filter_on_primary_author(self):
        """Verify person filter behavior.

        Such filtering should return packages whose primary
        content has an 'author' whose email matches the search term.
        """
        package_with_primary = package_factory(slug_key='test-pkg-api-pfopa')
        item_factory(primary_for=package_with_primary, author='author@a.co')

        # Test matches on primary content author
        response = self.client.get(PACKAGES_API_ENDPOINT, data={
            'format': 'json',
            'person': 'author@a.co'
        })
        json_response = json.loads(response.content)
        self.assertEqual(json_response['count'], 1)
        self.assertEqual(json_response['results'][0]['id'],
                         package_with_primary.pk)

    def test_person_filter_on_primary_editor(self):
        """Verify person filter behavior.

        Such filtering should return packages whose primary
        content has an 'editor' whose email matches the search term.
        """
        package_with_primary = package_factory(slug_key='test-pkg-api-pfope')
        item_factory(primary_for=package_with_primary, editor='editor@a.co')

        response = self.client.get(PACKAGES_API_ENDPOINT, data={
            'format': 'json',
            'person': 'editor@a.co'
        })
        json_response = json.loads(response.content)
        self.assertEqual(json_response['count'], 1)
        self.assertEqual(json_response['results'][0]['id'],
                         package_with_primary.pk)

    def test_person_filter_on_additional_author(self):
        """Verify person filter behavior.

        Such filtering should return packages whose additional
        content has an 'author' whose email matches the search term.
        """
        package_with_additional = package_factory(
            slug_key='test-pkg-api-pfoaa'
        )
        item_factory(primary_for=package_with_additional)
        item_factory(additional_for=package_with_additional,
                     author='author@b.co')

        response = self.client.get(PACKAGES_API_ENDPOINT, data={
            'format': 'json',
            'person': 'author@b.co'
        })
        json_response = json.loads(response.content)
        self.assertEqual(json_response['count'], 1)
        self.assertEqual(json_response['results'][0]['id'],
                         package_with_additional.pk)

    def test_person_filter_on_additional_editor(self):
        """Verify person filter behavior.

        Such filtering should return packages whose additional
        content has an 'editor' whose email matches the search term.
        """
        package_with_additional = package_factory(
            slug_key='test-pkg-api-pfoae'
        )
        item_factory(primary_for=package_with_additional)
        item_factory(additional_for=package_with_additional,
                     editor='editor@b.co')

        response = self.client.get(PACKAGES_API_ENDPOINT, data={
            'format': 'json',
            'person': 'editor@b.co'
        })
        json_response = json.loads(response.content)
        self.assertEqual(json_response['count'], 1)
        self.assertEqual(json_response['results'][0]['id'],
                         package_with_additional.pk)

    def test_has_primary_filter_on(self):
        """Verify has_primary filter behavior.

        Such filtering should exclude packages that don't have primary
        content items when the search value is 1.
        """
        package = package_factory(slug_key='test-pkg-api-hpf-on')
        response = self.client.get(PACKAGES_API_ENDPOINT, {
            'format': 'json',
            'has_primary': '0'
        })
        json_response = json.loads(response.content)
        self.assertEqual(json_response['count'], 1)
        self.assertEqual(json_response['results'][0]['id'], package.pk)

    def test_has_primary_filter_off(self):
        """Verify has_primary filter behavior.

        Such filtering should only include packages without primary
        content items when the search value is 0.
        """
        package_factory(slug_key='test-pkg-api-hpf-off')
        response = self.client.get(PACKAGES_API_ENDPOINT, {
            'format': 'json',
            'has_primary': '1'
        })
        json_response = json.loads(response.content)
        # Should only return the 5 items setup in setUpTestData
        self.assertEqual(json_response['count'], 5)

    def test_has_primary_filter_invalid(self):
        """Verify has_primary filter behavior.

        Such filtering should return an empty queryset when passed an
        invalid value.
        """
        package_factory(slug_key='test-pkg-api-hpf-inv')
        response = self.client.get(PACKAGES_API_ENDPOINT, {
            'format': 'json',
            'has_primary': 'x'
        })
        json_response = json.loads(response.content)
        self.assertEqual(json_response['count'], 0)


# @override_settings(ROOT_URLCONF='budget.test_urls')
class PrintPublicationAPITestCase(AuthedApiTestMixin, TestCase):
    multi_db = separate_db

    @classmethod
    def setUpTestData(cls):
        """Shared constructor behavior.

        Create multiple publications with multiple sections each, so
        that we can spot unnecessary queries on ForeignKey fields.
        """
        tdmn = PrintPublication.objects.create(name='The Dallas Morning News')
        PrintSection.objects.create(name='News', publication=tdmn)
        PrintSection.objects.create(name='Sports', publication=tdmn)
        PrintSection.objects.create(name='Inactive', publication=tdmn,
                                    is_active=False)
        al_dia = PrintPublication.objects.create(name='Al Dia')
        PrintSection.objects.create(name='News', publication=al_dia)
        PrintSection.objects.create(name='Sports', publication=al_dia)

    def test_query_count(self):
        """Verify prefetching behavior precludes additional queries.

        Extra queries shouldn't occur for foreign-keyed fields because
        the view should pre-fetch
        """
        with self.assertNumQueries(3, using=budget_db):
            self.client.get(PUBLICATION_API_ENDPOINT, data={'format': 'json'})

    def test_publication_active_filter(self):
        """Verify publication_active filter behavior.

        Such filtering should exclude/include publications based on
        their is_active value.
        """
        inactive = PrintPublication.objects.create(name='Inactive',
                                                   is_active=False)

        # We alredy have two active pubs from setUpTestData, so make sure they
        # show here
        response = self.client.get(PUBLICATION_API_ENDPOINT, {
            'format': 'json',
            'publication_active': '1'
        })
        json_response = json.loads(response.content)
        self.assertEqual(json_response['count'], 2)

        response = self.client.get(PUBLICATION_API_ENDPOINT, {
            'format': 'json',
            'publication_active': '0'
        })
        json_response = json.loads(response.content)
        self.assertEqual(json_response['count'], 1)
        self.assertEqual(json_response['results'][0]['id'], inactive.pk)


# @override_settings(ROOT_URLCONF='budget.test_urls')
class AuditTrailApiMixinTestCase(AuthedApiTestMixin, TestCase):
    """Ensure our mixin auto-adds user info & creates Change models."""
    multi_db = separate_db

    @classmethod
    def setUpTestData(cls):
        staff_factory()

    def setUp(self):
        self.creator = User.objects.create_user('user')
        self.client = APIClient()
        self.client.force_authenticate(user=self.creator)

    def test_user_saved_on_package_create(self):
        """Verify user info is stored on package save.

        Packages created via the API should have their created_by
        attribute set to the currently-authed user.
        """
        response = self.client.post(
            '%s?format=json' % PACKAGES_API_ENDPOINT,
            json.dumps({
                'hub': 'hub',
                'publishDate': [
                    get_timestamp_for_date('2016-08-18'),
                    get_timestamp_for_date('2016-08-19'),
                ],
                'slugKey': 'test-audit-usopc'
            }),
            content_type='application/json'
        )

        json_response = json.loads(response.content)
        self.assertEqual(response.status_code, 201)

        package = Package.objects.get(pk=json_response['id'])
        self.assertEqual(package.created_by, self.creator.pk)
        self.assertEqual(package.created_by_user, self.creator)

    def test_change_created_on_package_update(self):
        """Verify creation of Change model instance on package update.

        When a package is updated via the API, a change model should be
        created to reflect this modification.
        """
        package = package_factory(slug_key='test-audit-ccopu')

        response = self.client.patch(
            '%s%s/?format=json' % (PACKAGES_API_ENDPOINT, package.pk,),
            json.dumps({
                'publishedUrl': 'http://example.com/',
            }),
            content_type='application/json'
        )
        self.assertEqual(response.status_code, 200)

        change = Change.objects.get(
            item_id=package.pk,
            item_content_type='package'
        )
        self.assertEqual(change.package, package)
        self.assertEqual(change.by, self.creator.pk)

    def test_user_saved_on_item_create(self):
        """Verify user info is stored on item save.

        Item created via the API should have their created_by attribute
        set to the currently-authed user.
        """
        package = package_factory(slug_key='test-audit-usoic')

        response = self.client.post(
            '%s?format=json' % ITEMS_API_ENDPOINT,
            json.dumps({
                'hub': 'hub',
                'slugKey': 'slug',
                'budgetLine': 'Budget line',
                'authors': [{'name': 'Name'}],
                'primaryForPackage': package.pk
            }),
            content_type='application/json'
        )
        json_response = json.loads(response.content)
        self.assertEqual(response.status_code, 201)

        item = Item.objects.get(pk=json_response['id'])
        self.assertEqual(item.created_by, self.creator.pk)
        self.assertEqual(item.created_by_user, self.creator)

    def test_change_created_on_primary_item_update(self):
        """Verify creation of Change instance on primary item update.

        When a primary content item is updated via the API, a change
        model should be created to reflect this modification.
        """
        package = package_factory(slug_key='test-audit-ccopiu')
        item = item_factory(primary_for=package)

        response = self.client.patch(
            '%s%s/?format=json' % (ITEMS_API_ENDPOINT, item.pk,),
            json.dumps({
                'budgetLine': 'Updated budget line',
            }),
            content_type='application/json'
        )
        self.assertEqual(response.status_code, 200)

        change = Change.objects.get(
            item_id=item.pk,
            item_content_type='item'
        )
        self.assertEqual(change.package, package)
        self.assertEqual(change.by, self.creator.pk)

    def test_change_created_on_addl_item_update(self):
        """Verify creation of Change instance on additional item update.

        When an additional content item is updated via the API, a change
        model should be created to reflect this modification.
        """
        package = package_factory(slug_key='test-audit-ccoaiu')
        item_factory(primary_for=package)
        additional_item = item_factory(additional_for=package)

        response = self.client.patch(
            '%s%s/?format=json' % (ITEMS_API_ENDPOINT, additional_item.pk,),
            json.dumps({
                'budgetLine': 'Updated budget line',
            }),
            content_type='application/json'
        )
        self.assertEqual(response.status_code, 200)

        change = Change.objects.get(
            item_id=additional_item.pk,
            item_content_type='item'
        )
        self.assertEqual(change.package, package)
        self.assertEqual(change.by, self.creator.pk)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-06-24 02:39
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0008_package_publish_date'),
    ]

    operations = [
        migrations.RenameField(
            model_name='package',
            old_name='print_run_date',
            new_name='print_run_date_old',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-09-24 23:24
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0034_contentplacement_populate'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='package',
            name='is_print_placement_finalized',
        ),
        migrations.RemoveField(
            model_name='package',
            name='print_placements',
        ),
        migrations.RemoveField(
            model_name='package',
            name='print_run_date',
        ),
        migrations.RemoveField(
            model_name='package',
            name='print_run_date_old',
        ),
        migrations.RemoveField(
            model_name='package',
            name='print_section',
        ),
        migrations.RemoveField(
            model_name='package',
            name='print_system_slug',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-08-21 19:51
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0030_auto_20170820_2132'),
    ]

    operations = [
        migrations.AlterField(
            model_name='item',
            name='length',
            field=models.PositiveIntegerField(blank=True, help_text='How long is this article?', null=True, verbose_name='Length (in words)'),
        ),
        migrations.AlterField(
            model_name='item',
            name='slug_key',
            field=models.CharField(blank=True, help_text="Will be used, with package info, to generate additional items' slugs.", max_length=20, null=True),
        ),
        migrations.AlterField(
            model_name='item',
            name='type',
            field=models.CharField(choices=[('text', 'Text'), ('photo', 'Photo'), ('video', 'Video'), ('audio', 'Audio'), ('graphic', 'Graphic'), ('interactive', 'Interactive'), ('aggregation', 'Aggregation'), ('other', 'Other')], default='text', max_length=25),
        ),
        migrations.AlterField(
            model_name='package',
            name='headline_status',
            field=models.CharField(choices=[('drafting', 'Drafting'), ('voting', 'Voting'), ('finalized', 'Finalized')], default='drafting', max_length=9),
        ),
        migrations.AlterField(
            model_name='package',
            name='pub_date_resolution',
            field=models.CharField(blank=True, choices=[('m', 'Month'), ('w', 'Week'), ('d', 'Day'), ('t', 'Time')], default='d', editable=False, max_length=1, null=True),
        ),
        migrations.AlterField(
            model_name='package',
            name='slug_key',
            field=models.CharField(help_text='Miniature package descriptor. Will be used to generate the slug.', max_length=25, null=True),
        ),
        migrations.AlterField(
            model_name='printpublication',
            name='is_active',
            field=models.BooleanField(default=True, help_text='Should this publication be included in the current options users can choose from? If false, this publication will be archived, but it may be re-activated in the future.', verbose_name='Active?'),
        ),
        migrations.AlterField(
            model_name='printpublication',
            name='priority',
            field=models.PositiveSmallIntegerField(default=10, help_text='In the list of all active print publications, how high should this entry show up?'),
        ),
        migrations.AlterField(
            model_name='printsection',
            name='is_active',
            field=models.BooleanField(default=True, help_text='Should this publication be included in the current options users can choose from? If false, this publication will be archived, but it may be re-activated in the future.', verbose_name='Active?'),
        ),
        migrations.AlterField(
            model_name='printsection',
            name='priority',
            field=models.PositiveSmallIntegerField(default=10, help_text='In the list of all active print publications, how high should this entry show up?'),
        ),
    ]



# Imports from python.  # NOQA
from datetime import datetime
from datetime import timedelta
import os


# Imports from Django.
from django.core.management.base import BaseCommand
from django.db.models import F
from django.db.models import Func


# Imports from other dependencies.
import requests


# Imports from budget.
from budget.models import Package


TOKEN = os.environ.get("CUEBOT_TOKEN", None)

POST_URL = "http://apps.dallasnews.com/tools/cuebot/message/"

PAST_DUE_WARNINGS = {
    'first_warning': ' '.join([
        "Hello.",
        "You were an author or editor on the following package which had a",
        "planned publish date yesterday.",
        "If it published, would you kindly add the URL it published to?\n\n",
        "You can reply to me and say `{} was published to [story URL]`.\n\n",
        "If your package didn't publish, please delete the item or move its",
        "pub date.\n\n",
        "CueBot thanks you for keeping our newsroom budget tidy.\n",
    ]),
    'second_warning': ' '.join([
        "Hi. Your package is past due.",
        "If it published, please add the URL it published to.\n\n",
        "You can reply back and say `{} was published to [story URL]`.\n\n",
        "If it didn't publish, either delete it or move the pub date.\n\n",
        "CueBot thanks you for your compliance.\n",
    ]),
    'last_warning': ' '.join([
        "Citizen, your package is out of date.",
        "Please add publishing info, delete the package or change its",
        "pub date.\n\n",
        "Reply back with `{} was published to [story URL]`.\n\n",
        "CueBot thanks you for your compliance.\n",
    ]),
}


class Command(BaseCommand):
    help = "Reminds authors and editors of budget items past their deadline."

    def handle(self, *args, **options):

        past_due = datetime.combine(
            datetime.now().date(),
            datetime.min.time()
        )

        def get_recipients(package):
            if package.primary_content.editors:
                recipients = [
                    editor['email']
                    for editor in package.primary_content.editors
                ] + [
                    author['email']
                    for author in package.primary_content.authors
                ]
            else:
                recipients = [
                    author['email']
                    for author in package.primary_content.authors
                ]
            return recipients

        self.stdout.write("SENDING FIRST WARNINGS")
        # FIRST WARNING
        first_warn = past_due - timedelta(days=1)
        for package in Package.objects.annotate(
            publish_date_upper=Func(F('publish_date'), function='UPPER')
        ).filter(
            publish_date_upper__lt=past_due,
            publish_date_upper__gte=first_warn,
            published_url=None,
            primary_content__isnull=False
        ):
            recipients = get_recipients(package)
            requests.post(POST_URL, json={
                "token": TOKEN,
                "packageID": package.id,
                "recipients": recipients,
                "message": PAST_DUE_WARNINGS['first_warning'].format(
                    package.full_slug
                )
            })

            self.stdout.write(self.style.SUCCESS(
                "First warning sent for {} to {}".format(
                    package.full_slug,
                    ", ".join(recipients)
                )
            ))

        # SECOND WARNING: 2 days gone
        for package in Package.objects.annotate(
            publish_date_upper=Func(F('publish_date'), function='UPPER')
        ).filter(
            publish_date_upper__lt=first_warn,
            publish_date_upper__gte=first_warn - timedelta(days=1),
            published_url=None,
            primary_content__isnull=False
        ):
            recipients = get_recipients(package)
            requests.post(POST_URL, json={
                "token": TOKEN,
                "packageID": package.id,
                "recipients": recipients,
                "message": PAST_DUE_WARNINGS['second_warning'].format(
                    package.full_slug
                )
            })

            self.stdout.write(self.style.SUCCESS(
                "Second warning sent for {} to {}".format(
                    package.full_slug,
                    ", ".join(recipients)
                )
            ))

        # Last warning: 3 days gone
        for package in Package.objects.annotate(
            publish_date_upper=Func(F('publish_date'), function='UPPER')
        ).filter(
            publish_date_upper__lt=first_warn - timedelta(days=1),
            publish_date_upper__gte=first_warn - timedelta(days=2),
            published_url=None,
            primary_content__isnull=False
        ):
            recipients = get_recipients(package)
            requests.post(POST_URL, json={
                "token": TOKEN,
                "packageID": package.id,
                "recipients": recipients,
                "message": PAST_DUE_WARNINGS['last_warning'].format(
                    package.full_slug
                )
            })

            self.stdout.write(self.style.SUCCESS(
                "Last warning sent for {} to {}".format(
                    package.full_slug,
                    ", ".join(recipients)
                )
            ))
